/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include <cuda_runtime.h>
#include <cutil_inline.h>
#include "dAdq_cuda.h"

#define BLOCK_SIZE 8
#define BLOCKS(s) ((s)/BLOCK_SIZE + (((s)%BLOCK_SIZE) ? 1 : 0))

// all arrays column-minor
// Ns a multiple of BLOCK_SIZE, arrays padded with 0 as needed.
// testing: Nk the actual non-padded number of states

// gklt[k<Ns][l<Ns]
// dQdq[q<Nq][i<Ns][j<Ns]
// Aek [e<Nstim][k<Nk][i<Ns][j<Ns]
// dAdq_kl[q<Nq][k<Nk][l<Nk][i<Ns][j<Ns]
// dAdq[q<Nq][i<Ns][j<Ns]




extern "C" QUBCUDA_API void __cdecl qubcuda_init()
{
	char fake[5] = "fake";
  CUT_DEVICE_INIT(1, &fake);
}

extern "C" QUBCUDA_API void __cdecl qubcuda_done()
{
  CUDA_SAFE_CALL(cudaThreadExit());
}


extern "C" QUBCUDA_API dAdq_context* __cdecl dAdq_context_create(int Nstim, int Ns, int Nq)
{
  dAdq_context *ctx = new dAdq_context;
  ctx->Nstim = Nstim;
  ctx->Ns = Ns;
  ctx->Nq = Nq;
  ctx->Nb = BLOCKS(Ns);
  int Ns_full = ctx->Nb * BLOCK_SIZE;
  int Ns2 = Ns_full * Ns_full;
  int Nk2 = Ns * Ns;
  
  // set up cuda session?
  
  if (CUDA_SUCCESS != cudaMalloc(&ctx->gklt, Ns2*sizeof(float)))
	return NULL;
  if (CUDA_SUCCESS != cudaMalloc(&ctx->dQdq, Nq*Ns2*sizeof(float)))
	return NULL;
  if (CUDA_SUCCESS != cudaMalloc(&ctx->Aek, Nstim*Ns*Ns2*sizeof(float)))
	return NULL;
  if (CUDA_SUCCESS != cudaMalloc(&ctx->dAdq_kl, Nq*Nk2*Ns2*sizeof(float)))
	return NULL;
  if (CUDA_SUCCESS != cudaMalloc(&ctx->dAdq, Nq*Ns2*sizeof(float)))
	return NULL;
  
  dAdq_zero_kernel_call(Nstim, Ns_full, Ns, Nq, ctx->gklt, ctx->dQdq, ctx->Aek, ctx->dAdq_kl, ctx->dAdq);
  if ( cudaSuccess != cudaGetLastError() )
    return NULL;

  ctx->local = new float[max(Nstim*Ns_full, Nq) * Ns_full * Ns_full];

  return ctx;
}

extern "C" QUBCUDA_API void __cdecl dAdq_context_destroy(dAdq_context *ctx)
{
  if ( ! ctx )
    return;
  
  delete [] ctx->local;
  CUDA_SAFE_CALL(cudaFree(ctx->dAdq));
  CUDA_SAFE_CALL(cudaFree(ctx->dAdq_kl));
  CUDA_SAFE_CALL(cudaFree(ctx->Aek));
  CUDA_SAFE_CALL(cudaFree(ctx->dQdq));
  CUDA_SAFE_CALL(cudaFree(ctx->gklt));
  delete ctx;
}

extern "C" QUBCUDA_API void __cdecl dAdq_context_set_dQdq(dAdq_context *ctx, double ***dQdq)
{
  int Ns_full = ctx->Nb * BLOCK_SIZE;
  float *li = ctx->local;
  for (int q=0; q<ctx->Nq; ++q) {
    int i=0;
    for (; i<ctx->Ns; ++i) {
      int j=0;
      for (; j<ctx->Ns; ++j, ++li)
	*li = (float) dQdq[q][i][j];
      for (; j<Ns_full; ++j, ++li)
	*li = 0.0f;
    }
    for (; i<Ns_full; ++i)
      for (int j=0; j<Ns_full; ++j, ++li)
	*li = 0.0f;
  }
  
  CUDA_SAFE_CALL(cudaMemcpy(ctx->dQdq, ctx->local, ctx->Nq * Ns_full * Ns_full * sizeof(float),
			    cudaMemcpyHostToDevice));
}

extern "C" QUBCUDA_API void __cdecl dAdq_context_set_Aek(dAdq_context *ctx, double ****Aek)
{
  int Ns_full = ctx->Nb * BLOCK_SIZE;
  float *li = ctx->local;
  for (int e=0; e<ctx->Nstim; ++e) {

    for (int k=0; k<ctx->Ns; ++k) {
      int i=0;
      for (; i<ctx->Ns; ++i) {
	int j=0; 
	for (; j<ctx->Ns; ++j, ++li)
	  *li = (float) Aek[e][k][i][j];
	for (; j<Ns_full; ++j, ++li)
	  *li = 0.0f;
      }
      for (; i<Ns_full; ++i)
	for (int j=0; j<Ns_full; ++j, ++li)
	  *li = 0.0f;
    }

  }
  
  CUDA_SAFE_CALL(cudaMemcpy(ctx->Aek, ctx->local, ctx->Nstim * ctx->Ns * Ns_full * Ns_full * sizeof(float),
			    cudaMemcpyHostToDevice));
}

extern "C" QUBCUDA_API void __cdecl dAdq_context_set_gklt(dAdq_context *ctx, double **gklt)
{
  int Ns_full = ctx->Nb * BLOCK_SIZE;
  float *li = ctx->local;
  for (int i=0; i<ctx->Ns; ++i) {
    int j=0;
    for (; j<ctx->Ns; ++j, ++li)
      *li = (float) gklt[i][j];
    for (; j<Ns_full; ++j, ++li)
      *li = 0.0f;
  }
  
  CUDA_SAFE_CALL(cudaMemcpy(ctx->gklt, ctx->local, ctx->Ns * Ns_full * sizeof(float),
			    cudaMemcpyHostToDevice));
}

extern "C" QUBCUDA_API void __cdecl dAdq_context_get_dAdq(dAdq_context *ctx, double ***dAdq, int e)
{
  int Ns_full = ctx->Nb * BLOCK_SIZE;
    
  char buf[1024];
  cudaError_t err;

  float *Ak = ctx->Aek + e*ctx->Ns*Ns_full*Ns_full;
  dAdq_kl_kernel_call(Ns_full, ctx->Ns, ctx->Nq, ctx->gklt, ctx->dQdq, Ak, ctx->dAdq_kl);
    err = cudaGetLastError();                                    \
    if( cudaSuccess != err) {                                                \
        sprintf(buf, "Cuda error in file '%s' in line %i : %s.\n",    \
                __FILE__, __LINE__, cudaGetErrorString( err) );\
    }                                                                        \
    err = cudaThreadSynchronize();                                           \
    if( cudaSuccess != err) {                                                \
        sprintf(buf, "Cuda error in file '%s' in line %i : %s.\n",    \
                __FILE__, __LINE__, cudaGetErrorString( err) );\
    }                                                                        \

  dAdq_sum_kernel_call(Ns_full, ctx->Ns, ctx->Nq, ctx->dAdq_kl, ctx->dAdq);
    err = cudaGetLastError();                                    \
    if( cudaSuccess != err) {                                                \
        sprintf(buf, "Cuda error in file '%s' in line %i : %s.\n",    \
                __FILE__, __LINE__, cudaGetErrorString( err) );\
    }                                                                        \
    err = cudaThreadSynchronize();                                           \
    if( cudaSuccess != err) {                                                \
        sprintf(buf, "Cuda error in file '%s' in line %i : %s.\n",    \
                __FILE__, __LINE__, cudaGetErrorString( err) );\
    }                                                                        \
  
  CUDA_SAFE_CALL(cudaMemcpy(ctx->local, ctx->dAdq, ctx->Nq * Ns_full * Ns_full * sizeof(float),
			    cudaMemcpyDeviceToHost));
  
  float *li = ctx->local;
  for (int q=0; q<ctx->Nq; ++q) {
    for (int i=0; i<ctx->Ns; ++i) {
      for (int j=0; j<ctx->Ns; ++j, ++li)
	dAdq[q][i][j] = (double) *li;
      li += (Ns_full - ctx->Ns);
    }
    li += (Ns_full - ctx->Ns) * Ns_full;
  }
}

